--==================================================================================================--
--	chataclysm (noun)
--	1) a disconnect that results in a user losing all their chat channels 
--	2) the name of an addon which automatically repairs the damage caused by a chataclysmic event
--	source: The Medivh-Wrynn Common Dictionary (Revised Edition)
--==================================================================================================--

Chataclysm = LibStub("AceAddon-3.0"):NewAddon("Chataclysm", "AceEvent-3.0", "AceTimer-3.0")
local Chataclysm = Chataclysm

-- chat events
local CA = {
	"INIT",
	"DONE",
	"ADD",
	"REMOVE",
	"BUFFER",
	"ADD_BUFFER",
	"REMOVE_BUFFER",
	"PLACEHOLDER",
	"ADD_PLACEHOLDER",
	"REMOVE_PLACEHOLDER",
}

local DL = {
	[0] = "none",
	"error",
	"test",
	"config",
	"events",
	"spam",
}
local dbgprint = { squelch = function(...) return nil end }
function Chataclysm:SetDebugLevel(iDL)
	for i = 0,     iDL do dbgprint[ DL[i] ] = print; end
	for i = iDL+1, #DL do dbgprint[ DL[i] ] = dbgprint.squelch; end
end

local defaults = {
	global = {
		debugLevel = 0,        -- DL setting for verbosity of debug output 
		initialDelay = 20,     -- number of seconds until we will check/fix after addon load unless we see a CHANNEL_UI_UPDATE event 
		incrementalDelay = 2,  -- number of seconds until we will check/fix after seeing a CHANNEL_UI_UPDATE event unless we see another CHANNEL_UI_UPDATE event 
	},
	profile = {
		channelNames = {}, -- key = channel number (int), value = channel name (string)
		frameList = {},    -- key = channel name (string), value = table { key = (int), value = name of chat frame (string) }
	},
}

Chataclysm.frames = {} -- key = name of chat frame (string), value = table { "frame" = the chatframe itself (table), "hasChannel" = table { key = channel name (string), value = true (if exists) } }
Chataclysm.channels = {} -- key = id of channel (int), value = current name of the channel (generalized, no zone)
Chataclysm.actions = {} -- key = id of channel (int), value = table { "type" = action scheduled to happen to channel (CA), "name" = channel we want to see in this slot }
Chataclysm.serverChannels = {}  -- key = id of channel (int), value = generalized name without zone decoration (string)

function Chataclysm:PrintActions()
	for iChannel=1, 10 do
		print(iChannel, self.channels[iChannel], CA[self.actions[iChannel].type], self.actions[iChannel].name)
	end
end

function Chataclysm:GetCurrentFrames()
	self.frames = {}

	local function GetChannelList(...)
		local tbl = {}
		for iOffset = 1, select("#", ...), 2 do
			tbl[select(iOffset, ...)] = true
		end
		return tbl;
	end

	for i=1, NUM_CHAT_WINDOWS do
		local tblChatFrame = _G["ChatFrame" .. tostring(i)]
		if (tblChatFrame ~= nil) then
			local channels = GetChannelList(GetChatWindowChannels(tblChatFrame:GetID()))
			self.frames[tblChatFrame.name] = { frame = tblChatFrame, hasChannel = channels }
		end
	end

	for k,v in pairs(self.frames) do
		dbgprint.spam(k, v.frame:GetID())
		for n,_ in pairs(v.hasChannel) do
			dbgprint.spam("  ", n)
		end
	end
end

function Chataclysm:GetCurrentChannels()
	self.serverChannels = {}
	for _,v in pairs( {EnumerateServerChannels()} ) do
		self.serverChannels[GetChannelName(v)] = v
	end
	self.serverChannels[0] = nil
	for k,v in pairs(self.serverChannels) do dbgprint.spam(k,v) end

	self.channels = {}
	for i=1, 10 do
		self.channels[i] = self.serverChannels[i] or select(2, GetChannelName(i))
		dbgprint.spam(i, self.channels[i])
	end
end

local strTempChannelPrefix = "ChataclysmPlaceholder"
local strTempChannelFormat = ""
local function IsTempChannel(name) return strfind(name, strTempChannelPrefix, 1, true) end

-- Called when the addon is loaded
function Chataclysm:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("ChataclysmDB", defaults)
	self.db.RegisterCallback(self, "OnDatabaseShutdown", "SaveChatConfig")
	
	for k,v in pairs(CA) do CA[v] = k end
	for k,v in pairs(DL) do DL[v] = k end
	self:SetDebugLevel(self.db.global.debugLevel)
	
	self.noticeHandler = "DelayFirstChatCheck"
	self.noticeUserHandler = nil
	self.zoneChangedHandler = nil

	local seed = math.modf(mod((GetTime() * 1000), 100000))
	strTempChannelFormat = format("%s%%i%i", strTempChannelPrefix, (seed))
	dbgprint.test(strTempChannelFormat)

	-- init actions struct
	for i=1, 10 do
		self.actions[i] = { type = CA.INIT, name = self.db.profile.channelNames[i] }
	end
end

-- Called when the addon is enabled
function Chataclysm:OnEnable()
	self:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE", self.noticeHandler)
	
	self.timer = self:ScheduleTimer("FirstChatCheck", self.db.global.initialDelay)
end

-- Called when the addon is Disabled
function Chataclysm:OnDisable()
	if self.noticeHandler      then self:UnregisterEvent("CHAT_MSG_CHANNEL_NOTICE")      end
	if self.noticeUserHandler  then self:UnregisterEvent("CHAT_MSG_CHANNEL_NOTICE_USER") end
	if self.zoneChangedHandler then self:UnregisterEvent("ZONE_CHANGED_NEW_AREA")        end
end

function Chataclysm:EventHandler(...) dbgprint.events(...) end

function Chataclysm:DelayFirstChatCheck(event, message, _, _, _, _, _, _, channelNumber, channelName, ...)
	dbgprint.events("DelayFirstChatCheck", event, message, channelNumber, channelName)
	if self.timer then
		self:CancelTimer(self.timer, false)
		self.timer = self:ScheduleTimer("FirstChatCheck", self.db.global.incrementalDelay)
	end
end

function Chataclysm:FirstChatCheck()
	self.timer = nil

	self.noticeHandler = "ResolveActionNotice"
	self:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE", self.noticeHandler)
	self.noticeUserHandler = "ResolveActionNoticeUser"
	self:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE_USER",  self.noticeUserHandler)

	self:CheckChat()
end

function Chataclysm:RetryPlaceholders(...)
	dbgprint.events("RetryPlaceholders", ...)
	for iChannel = 1, 10 do
		local tblAction = self.actions[iChannel] or {}
		if (tblAction.type == CA.PLACEHOLDER) then
			tblAction.type = CA.REMOVE_PLACEHOLDER
		end
	end

	self:ScheduleTimer("CheckChat", 5)
end


function Chataclysm:ResolveActionNoticeUser(event, message, _, _, _, _, _, zoneChannel, channelNumber, channelName, ...)
	dbgprint.events("ResolveActionNoticeUser", event, message, zoneChannel, channelNumber, channelName)
	local tblAction = self.actions[channelNumber] or {}
	if (message == "INVALID_NAME" and tblAction.type == CA.ADD and zoneChannel ~= 0) then
		dbgprint.config("creating placeholder channel for", channelNumber, tblAction.name)
		tblAction.type = CA.ADD_PLACEHOLDER
		
		if (not self.zoneChangedHandler) then
			self.zoneChangedHandler = "RetryPlaceholders"
			self:RegisterEvent("ZONE_CHANGED_NEW_AREA", self.zoneChangedHandler)
		end
		
		self:ScheduleTimer("CheckChat", 0.3)
	end
end

function Chataclysm:ResolveActionNotice(event, message, _, _, _, _, _, _, channelNumber, channelName, ...)
	dbgprint.events("ResolveActionNotice", event, message, channelNumber, channelName)
	local tblAction = self.actions[channelNumber] or {}
	
	if (tblAction.type == CA.DONE) then
		return
	end

	if (message == "YOU_LEFT") then
		if (tblAction.type ~= CA.REMOVE or tblAction.type ~= CA.REMOVE_PLACEHOLDER) then
			tblAction.type = CA.INIT
		elseif (tblAction.type ~= CA.REMOVE_BUFFER) then
			tblAction.type = CA.DONE
		else
			dbgprint.error("left channel for unexpected reason", channelNumber, channelName)
		end
		self:ScheduleTimer("CheckChat", 0.3)

	elseif (message == "YOU_JOINED") then
		if (tblAction.type == CA.ADD and (channelName == tblAction.name or channelNumber == GetChannelName(tblAction.name))) then
			tblAction.type = CA.DONE
		elseif (tblAction.type == CA.ADD_BUFFER and IsTempChannel(channelName)) then
			tblAction.type = CA.BUFFER
		elseif (tblAction.type == CA.ADD_PLACEHOLDER and IsTempChannel(channelName)) then
			tblAction.type = CA.PLACEHOLDER
		else
			dbgprint.error("ERROR: channel joined but name does not match expected", channelNumber, channelName)
			tblAction.type = CA.DONE -- using CA.INIT is likely to create loops
		end
		self:CheckChat()
	end
end

function Chataclysm:CheckChat()
	self:UpdateActionTable()
	if (self:FExecuteNextAction()) then 
		return 
	end
	
	-- if we had no joins or leaves then we can remove any BUFFER channels added for padding
	dbgprint.config("removing any buffers")
	for iSaved=1, 10 do
		if (self.actions[iSaved].type == CA.BUFFER) then 
			self.actions[iSaved].type = CA.REMOVE_BUFFER
		end
	end
	if (self:FExecuteNextAction()) then
		return
	end
	
	-- if there are no placeholder channels then we've finished the task of checking channels and can stop monitoring events
	for iSaved=1, 10 do
		if (self.actions[iSaved].type == CA.PLACEHOLDER) then 
			return
		end
	end

	dbgprint.config("removing event handlers")
	if self.noticeHandler      then self:UnregisterEvent("CHAT_MSG_CHANNEL_NOTICE")      end
	if self.noticeUserHandler  then self:UnregisterEvent("CHAT_MSG_CHANNEL_NOTICE_USER") end
	if self.zoneChangedHandler then self:UnregisterEvent("ZONE_CHANGED_NEW_AREA")        end
	self.noticeHandler      = nil
	self.noticeUserHandler  = nil
	self.zoneChangedHandler = nil

end


-- returns true if it made a call to add or remove a channel
-- when adding a channel it also repairs associations with chat windows
function Chataclysm:FExecuteNextAction()
	for iChannel=1, 10 do
		if (self.actions[iChannel].type == CA.ADD) then
			local strChannel = self.actions[iChannel].name
			JoinPermanentChannel(strChannel)

			-- channel should be in the right place, now make sure its going to the correct windows
			for _,strSavedFrame in pairs(self.db.profile.frameList[strChannel]) do
				local tblChatFrame = self.frames[strSavedFrame]
				if (tblChatFrame == nil) then
					dbgprint.error(iChannel, strChannel, "cannot be added to chat window", strSavedFrame, "as it no longer exists")
				elseif (not tblChatFrame.hasChannel[strSaved]) then
					dbgprint.spam("adding channel", strChannel, "to chat window", strSavedFrame)
					ChatFrame_AddChannel(tblChatFrame.frame, strChannel)
				end
			end

			return true;

		elseif (self.actions[iChannel].type == CA.ADD_BUFFER or 
		        self.actions[iChannel].type == CA.ADD_PLACEHOLDER ) then
			JoinTemporaryChannel(format(strTempChannelFormat, iChannel))
			return true;

		elseif (self.actions[iChannel].type == CA.REMOVE or 
		        self.actions[iChannel].type == CA.REMOVE_BUFFER or  
		        self.actions[iChannel].type == CA.REMOVE_PLACEHOLDER ) then
			LeaveChannelByName(self.channels[iChannel])
			return true;
		end
	end		
end
		

function Chataclysm:FEnsureSlotting(index)
	-- recursively ensure that there are channels below the slot we want to fill, 
	-- filling them with a temp channel if need be
	index = index - 1
	
	if (index == 0) then 
		return true;
	end
	
	-- if there are remove operations scheduled we should wait for those to finish 
	-- processing before trying to add new channels to slots after them
	local tblAction = self.actions[index] or {}
	if (tblAction.type == CA.REMOVE or 
	    tblAction.type == CA.REMOVE_BUFFER or 
	    tblAction.type == CA.REMOVE_PLACEHOLDER ) then 
		return false;
	end
	
	if (tblAction.type == CA.DONE and not self.channels[index]) then
		self.actions[index].type = CA.ADD_BUFFER
	end
	
	return self:FEnsureSlotting(index)
end

function Chataclysm:UpdateActionTable()
	self:GetCurrentFrames()
	self:GetCurrentChannels()

	for iSaved=1, 10 do
		-- we only need evaluate channels that are still in the INIT action state
		if (self.actions[iSaved].type == CA.INIT) then
			local strSaved = self.actions[iSaved].name
			local iActual = (strSaved ~= nil) and GetChannelName(strSaved) or iSaved

			if (iActual == iSaved) then
				-- index matches saved channel (or is empty if no channel was saved), mark it DONE dude
				self.actions[iSaved].type = CA.DONE
			else
				dbgprint.error("fix needed for", iSaved, strSaved)
				
				-- are we in the wrong slot? (as opposed to just absent)
				if (iActual ~= 0) then
					dbgprint.config("  ", strSaved, "was found at slot", iActual, "and needs to be cleared")
					self.actions[iActual].type = CA.REMOVE
				end				
				
				-- is there another channel sitting in our spot?
				if (self.channels[iSaved] ~= nil) then
					dbgprint.config("  ", self.channels[iSaved], "is sitting in our place - blow it away")
					-- TODO REVIEW should we save removed channels and append after fixing?
					self.actions[iSaved].type = CA.REMOVE
				end

				-- not safe to add if there are removals in the works
				if (iActual == 0 and self.channels[iSaved] == nil and self:FEnsureSlotting(iSaved)) then
					-- dbgprint.config("adding channel", strSaved)
					self.actions[iSaved].type = CA.ADD
				end
			end
		end
	end
end


function Chataclysm:SaveChatConfig()
	self:GetCurrentChannels()
	
	for i=1, 10 do
		local name = self.channels[i]
		
		-- if we haven't been able to replace a placeholder channel yet, then make sure
		-- the channel we were holding a spot for is saved out for next session
		if (self.actions[i].type == CA.PLACEHOLDER) then
			dbgprint.config("placeholder channel flag still set for", i, name)
			if (IsTempChannel(name)) then
				name = self.actions[i].name
			end
		-- if a buffer channel is somehow still there, then just nuke it
		elseif (self.actions[i].type == CA.BUFFER) then
			dbgprint.error("buffer channel flag still set for", i, name)
			if (IsTempChannel(name)) then
				name = nil
			end
		end
			
		self.db.profile.channelNames[i] = name
		if (name ~= nil) then
			self.db.profile.frameList[name] = {}
		end
	end


	self:GetCurrentFrames()
	for frameName,v in pairs(self.frames) do
		for channelName,_ in pairs(v.hasChannel) do
			tinsert(self.db.profile.frameList[channelName], frameName)
		end
	end
	
end

